@@ -1,5 +1,6 @@
 /* 
 **  mod_mp3.c
+**  $Id: mod_mp3.c,v 1.118 2002/05/20 00:45:56 brian Exp $
 */ 
 
 #include "mod_mp3.h"
@@ -11,11 +12,23 @@ extern mp3_dispatch internal;
 #ifdef MYSQL_ENABLED
 extern mp3_dispatch mysql;
 #endif
+#ifdef PGSQL_ENABLED
+extern mp3_dispatch pgsql;
+#endif
+#ifdef PLAYLIST_ENABLED
+extern mp3_dispatch playlist;
+#endif
 
 mp3_dispatch *mp3_dispatches[] = {
 	&internal,
 #ifdef MYSQL_ENABLED
 	&mysql,
+#endif
+#ifdef PGSQL_ENABLED
+	&pgsql,
+#endif
+#ifdef PLAYLIST_ENABLED
+	&playlist,
 #endif
 	NULL,
 };
@@ -25,17 +38,17 @@ void send_headers(request_rec *r, mp3_conf* cfg, request_data *request) {
 #ifdef DEBUG
 	printf("TYPE %d\n", request->type);
 #endif
-	if (request->type == 1) {
+	if (request->type == ICESTREAM) {
 		send_icecast_headers(r, cfg, request);
-	} else if (request->type == 0){
+	} else if (request->type == SHOUTSTREAM){
 		send_shout_headers(r, cfg, request);
-	} else if (request->type == 2){
+	} else if (request->type == VORBISSTREAM){
 		send_ogg_headers(r, cfg, request);
 	} else {
+		/* Basically we don't know, or we don't support */
 		r->content_type = cfg->content_type;
 		ap_send_http_header(r);
 	}
-
 }
 
 /* Creates the per virtualhost, directory, location information */
@@ -47,8 +60,8 @@ static void *mconfig_for_directory(pool *p, char *dir) {
 	cfg->limit = DEFAULT_LIMIT;
 	cfg->random_enabled = 0;
 	cfg->cache_enabled = 0;
+	cfg->max_bytes = 0;
 	cfg->encoder = NULL;
-	cfg->files = NULL;
 	cfg->stream = NULL;
 	cfg->cast_name = DEFAULT_CAST_NAME;
 	cfg->genre_name = DEFAULT_GENRE_NAME;
@@ -65,22 +78,32 @@ static void *mconfig_for_directory(pool *p, char *dir) {
 static void *mconfig_for_server(pool *p, server_rec *s) {
 	mp3_server_conf *scfg = ap_pcalloc (p, sizeof (mp3_server_conf));
 	int shmid = -1;
-	scoreboard *shm = NULL;
+	mp3_scoreboard *shm = NULL;
 
-	if ((shmid = shmget(shmkey, sizeof(scoreboard), IPC_CREAT | 0666)) < 0) {
-		printf("%s - Cannot create Shared Memory", __FILE__);
+	//if ((shmid = shmget(shmkey, sizeof(scoreboard), IPC_CREAT | 0600 )) < 0) {
+	if ((shmid = shmget(shmkey, sizeof(mp3_scoreboard), IPC_CREAT | SHM_R | SHM_W )) < 0) {
+#ifdef LINUX
+		if (errno == ENOSYS) {
+				ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_EMERG, s,
+					"Your kernel was built without CONFIG_SYSVIPC\n"
+					"Please consult the Apache FAQ for details");
+		} 
+#endif 
+
+		printf("%s - Cannot create Shared Memory: %s(%d)(%d)\n", __FILE__, strerror(errno),errno, shmid);
 		exit(1);
 	}
 
 	if (!(shm = get_scoreboard(shmid))) {
-		printf("%s - Cannot attach to Shared Memory", __FILE__);
+		printf("%s - Cannot attach to Shared Memory: %s(%d)\n", __FILE__, strerror(errno),errno);
 		exit(1);
 	}
+//	shmctl(shmid, IPC_RMID, 0);
 
 	ap_register_cleanup(p, scfg, cleanup_scoreboard, ap_null_cleanup);
 
 	scfg->shmid = shmid;
-	memset(shm, 0, sizeof(scoreboard));
+	memset(shm, 0, sizeof(mp3_scoreboard));
 	shm->time = time(NULL);
 	scfg->generation = time(NULL);
 	scfg->board = shm;
@@ -92,9 +115,11 @@ static int stream_content(request_rec *r, mp3_conf *cfg, mp3_data *content, requ
 	FILE *file = NULL;
 	int b = 0;
 	int x = 0;
+	int bytes_to_send = 0;
 	int latch = 0;
 	unsigned char *temp = NULL;
 	const char *message = NULL;
+	mp3_server_conf *scfg = ap_get_module_config(r->server->module_config, &mp3_module);
 
 #ifdef DEBUG
 		printf ("Playing %s \n", content->name);
@@ -112,16 +137,25 @@ static int stream_content(request_rec *r, mp3_conf *cfg, mp3_data *content, requ
 
 	ap_hard_timeout("mod_mp3_write", r);
 
+	connection_set_file(r, scfg, content->signature, content->name);
+
 	if (content->data) {
+		/* Check for bytes per track limit */   
+		if (cfg->max_bytes > 0 && cfg->max_bytes > content->size)
+			bytes_to_send = cfg->max_bytes;
+		else
+			bytes_to_send = content->size;
+
+	
 		if (request->shout) {
 /* Bug in this at the moment */
 			temp = (unsigned char *)content->data;
-			for(x = 0; x < content->size; x++) {
+			for(x = 0; x < bytes_to_send; x++) {
 				if (shout_write(r, temp[x], content->name, content->artist, request->url, &latch) == -1)
 					return HTTP_REQUEST_TIME_OUT;
 			}
 		} else {
-			x = ap_send_mmap(content->data, r, 0, content->size);
+			x = ap_send_mmap(content->data, r, 0, bytes_to_send);
 			if (!x)
 				return HTTP_REQUEST_TIME_OUT;
 		} 
@@ -152,6 +186,15 @@ static int stream_content(request_rec *r, mp3_conf *cfg, mp3_data *content, requ
 	return OK;
 }
 
+static void mp3_child_init(server_rec *s, pool *p) {
+	/*
+	mp3_server_conf *scfg = ap_get_module_config(r->server->module_config, &mp3_module);
+	if (scfg->directory_server) {
+	}
+	*/
+}
+
+
 
 #ifdef SELECT_ENABLED
 static int mp3_selection_handler(request_rec *r) {
@@ -178,7 +221,7 @@ static int mp3_selection_handler(request_rec *r) {
 
 	ap_rprintf(r, "<FORM ACTION=\"http://%s:%d%s\" METHOD=\"GET\">\n", r->hostname, r->server->port, r->uri);
 	ap_rputs("<TABLE>\n", r);
-	while ((bank = cfg->dispatch->each(cfg->context, r->pool, info->files, info->token))) {
+	while ((bank = cfg->dispatch->each(cfg->context, r->pool, info->files, info->token, info->random))) {
 		x++;
 		if (x % 2)
 			ap_rputs("<TR><TD BGCOLOR=\"#ffffff\">\n", r);
@@ -202,6 +245,7 @@ static int mp3_selection_handler(request_rec *r) {
 	ap_rputs("<INPUT TYPE=\"reset\" VALUE=\"Reset\">\n", r);
 	ap_rputs("<INPUT TYPE=\"submit\" NAME=\"submit\" VALUE=\"Submit\">\n", r);
 	ap_rputs("</FORM>\n", r);
+	ap_rputs("</BODY></HTML>\n", r);
 
 	return OK;
 }
@@ -282,7 +326,7 @@ static int mp3_rss_handler(request_rec *r) {
 	print_channel(r, cfg);
 
 /* Brian, do pattern */
-	while ((bank = cfg->dispatch->each(cfg->context, r->pool, info->files, info->token))) {
+	while ((bank = cfg->dispatch->each(cfg->context, r->pool, info->files, info->token, info->random))) {
 		ap_rputs("<item>\n", r);
 		ap_rprintf(r, "<title>%s</title>\n", escape_xml(r->pool, bank->name));
 		ap_rprintf(r, "<link>http://%s:%d%s?op=play&amp;song=%s", r->hostname, r->server->port, r->uri, bank->signature);
@@ -321,7 +365,7 @@ static int mp3_rdf_handler(request_rec *r) {
 	print_channel_rdf(r, cfg, files);
 
 /* Brian, do pattern */
-	while ((bank = cfg->dispatch->each(cfg->context, r->pool, files, info->token))) {
+	while ((bank = cfg->dispatch->each(cfg->context, r->pool, files, info->token, info->random))) {
 		ap_rprintf(r, "\t<item rdf:about=\"http://%s:%d%s?op=play&amp;song=%s\">\n", r->hostname, r->server->port, r->uri, bank->signature);
 		ap_rprintf(r, "\t\t<title>%s</title>\n", escape_xml(r->pool, bank->name));
 		ap_rprintf(r, "\t\t<link>http://%s:%d%s?op=play&amp;song=%s", r->hostname, r->server->port, r->uri, bank->signature);
@@ -378,7 +422,7 @@ static int mp3_mbm_handler(request_rec *r) {
 	print_channel_mbm(r, cfg, files);
 
 /* Brian, do pattern */
-	while ((bank = cfg->dispatch->each(cfg->context, r->pool, info->files, info->token))) {
+	while ((bank = cfg->dispatch->each(cfg->context, r->pool, info->files, info->token, info->random))) {
 		ap_rprintf(r, "\t<mm:Track rdf:about=\"http://%s:%d%s?op=play&amp;song=%s\">\n", r->hostname, r->server->port, r->uri, bank->signature);
 		ap_rprintf(r, "\t\t<dc:title>%s</dc:title>\n", escape_xml(r->pool, bank->name));
 		if (bank->artist) 
@@ -430,17 +474,17 @@ static int mp3_pls_handler(request_rec *r) {
 		version = 2;
 
 	if (version)
-		ap_rprintf(r, "numberofentries=%d\n", cfg->files->nelts);
+		ap_rprintf(r, "numberofentries=%d\n", cfg->dispatch->count(cfg->context, r->pool, info->files, info->token));
 
-	while ((bank = cfg->dispatch->each(cfg->context, r->pool, info->files, info->token))) {
+	while ((bank = cfg->dispatch->each(cfg->context, r->pool, info->files, info->token, info->random))) {
 		x++;
 		if (version) {
-			ap_rprintf(r, "File%d=http://%s:%d%s?op=play&song=%s",x+1 , r->hostname, r->server->port, r->uri, bank->signature);
+			ap_rprintf(r, "File%d=http://%s:%d%s?op=play&song=%s",x , r->hostname, r->server->port, r->uri, bank->signature);
 			if (info->type == 2)
 				ap_rputs("&type=.ogg", r);
 			ap_rputs("\n", r);
-			ap_rprintf(r, "Title%d=%s\n", x+1, bank->name);
-			ap_rprintf(r, "Length%d=-1\n", x+1);
+			ap_rprintf(r, "Title%d=%s\n", x, bank->name);
+			ap_rprintf(r, "Length%d=-1\n", x);
 		} else {
 			ap_rprintf(r, "http://%s:%d%s?op=play&song=%s", r->hostname, r->server->port, r->uri, bank->signature);
 			if (info->type == 2)
@@ -477,7 +521,7 @@ static int mp3_m3u_handler(request_rec *r) {
 	if (version)
 		ap_rprintf(r, "#EXTM3U\n");
 
-	while ((bank = cfg->dispatch->each(cfg->context, r->pool, info->files, info->token))) {
+	while ((bank = cfg->dispatch->each(cfg->context, r->pool, info->files, info->token, info->random))) {
 			/* 1st arg is playlength in secs; -1 for "don't know" */
 		if (version) {
 			if (bank->artist)
@@ -533,7 +577,7 @@ request_data * create_request(request_rec *r, mp3_conf *cfg) {
 	request->args = NULL;
 	request->udp = 0;
 	request->shout = 0;
-	request->type = -1;
+	request->type = HTTPSTREAM;
 	request->random = cfg->random_enabled;
 	request->limit = cfg->limit ? cfg->limit : DEFAULT_LIMIT;
 	request->files = NULL;
@@ -557,13 +601,13 @@ request_data * create_request(request_rec *r, mp3_conf *cfg) {
 	}
 
 	if (request->shout) {
-		request->type = 0;
+		request->type = SHOUTSTREAM;
 	} else if (request->udp) {
-		request->type = 1;
+		request->type = ICESTREAM;
 	} else if (!mp3_match(cfg->content_type, "audio/x-ogg")) {
-		request->type = 2;
+		request->type = VORBISSTREAM;
 	} else {
-		request->type = -1;
+		request->type = HTTPSTREAM;
 	}
 
 	if (r->args) {
@@ -578,11 +622,11 @@ request_data * create_request(request_rec *r, mp3_conf *cfg) {
 		/* I question if this is still needed. -Brian */
 		if ((type = ap_table_get(request->args, "type"))) {
 			if (!mp3_match(".ogg", type)) {
-				request->type = 2;
+				request->type = VORBISSTREAM;
 			} 
 		}
 		if (!mp3_match("audio/x-ogg", cfg->content_type)) {
-			request->type = 2;
+			request->type = VORBISSTREAM;
 		} 
 		if (request->pattern) {
 			request->files = cfg->dispatch->search(cfg->context, r->pool, request->pattern, request->limit);
@@ -593,7 +637,7 @@ request_data * create_request(request_rec *r, mp3_conf *cfg) {
 #ifdef CRAP
 	/* Just to solve a bug */
 	if (cfg->cache_enabled) {
-		request->type = -1;
+		request->type = HTTPSTREAM;
 	}
 #endif
 	if (request->order && (!mp3_match(request->order, "random"))) {
@@ -612,11 +656,45 @@ request_data * create_request(request_rec *r, mp3_conf *cfg) {
 static int mp3_request_init(request_rec *r) {
 	mp3_conf *cfg = ap_get_module_config(r->per_dir_config, &mp3_module);
 	if (cfg->dispatch->request)
-		cfg->dispatch->request(cfg->context, r->pool);
+		cfg->dispatch->request(cfg->context, cfg, r->pool);
 
 	return DECLINED;
 }
 
+static int mp3_status_handler(request_rec *r) {
+	mp3_server_conf *scfg = ap_get_module_config(r->server->module_config, &mp3_module);
+	int x = 0;
+
+	r->content_type = "text/html";      
+	ap_send_http_header(r);
+	ap_rprintf(r, "%s<HTML><HEAD>\n<TITLE>mod_mp3 status</TITLE>\n</HEAD><BODY><TABLE><TR>\n", DOCTYPE_HTML_3_2);
+	ap_rprintf(r, "<TD>#</TD><TD>stream type</TD><TD>Connecting Host</TD><TD>Signature of file being sent</TD><TD>Title of file</TD></TR>\n");
+	for (x = 0; x < CHILDREN; x++) {
+		if (scfg->board->servers[x].status) {
+			ap_rprintf(r, "<TR><TD>%d</TD>",x); 
+			ap_rputs("<TD>", r);
+			if (scfg->board->servers[x].type == HTTPSTREAM) {
+				ap_rprintf(r, "HTTP stream\t");
+			} else if (scfg->board->servers[x].type == ICESTREAM) {
+				ap_rprintf(r, "Ice Stream\t");
+			} else if (scfg->board->servers[x].type == SHOUTSTREAM) {
+				ap_rprintf(r, "Shout stream\t");
+			} else if (scfg->board->servers[x].type == VORBISSTREAM) {
+				ap_rprintf(r, "Ogg Vorbis stream\t");
+			} else {
+				ap_rprintf(r, "unknown\t");
+			}
+			ap_rputs("</TD>", r);
+			ap_rprintf(r, "<TD>%s</TD><TD>%s</TD><TD>%s</TD></TR>\n", 
+				scfg->board->servers[x].remote_ip,
+				scfg->board->servers[x].current_file, scfg->board->servers[x].title);
+		}
+	}
+	ap_rputs("</TR></TABLE></BODY></HTML>\n", r);
+
+	return OK;
+}
+
 static int mp3_fixup(request_rec *r) {
 	mp3_conf *cfg = ap_get_module_config(r->per_dir_config, &mp3_module);
 	request_data *info = NULL;
@@ -684,7 +762,7 @@ static int mp3_play_handler(request_rec *r) {
 	request_data *info = ap_get_module_config(r->request_config, &mp3_module);
 
  
-	if ((status = register_connection(r, scfg, cfg->connection_limit, -1)) != OK) {
+	if ((status = register_connection(r, scfg, cfg->connection_limit, info->type)) != OK) {
 		return status;
 	}
 #ifdef DEBUG
@@ -693,24 +771,15 @@ static int mp3_play_handler(request_rec *r) {
 
 	send_headers(r, cfg, info);
 
-	if (info->random) {
-		do {
-			while ((bank = cfg->dispatch->random(cfg->context, r->pool))) {
-				if (stream_content(r, cfg, bank, info) != OK)
-					return OK;
-			}
-		} while (cfg->loop);
-	} else {
-		do {
+	do {
 #ifdef DEBUG
-				printf("Using dispatch %s\n", cfg->dispatch->name);
+		printf("Using dispatch %s\n", cfg->dispatch->name);
 #endif
-			while ((bank = cfg->dispatch->each(cfg->context, r->pool, info->files, info->token))) {
-				if (stream_content(r, cfg, bank, info) != OK)
-					return OK;
-			}
-		} while (cfg->loop);
-	}
+		while ((bank = cfg->dispatch->each(cfg->context, r->pool, info->files, info->token, info->random))) {
+			if (stream_content(r, cfg, bank, info) != OK)
+				return OK;
+		}
+	} while (cfg->loop);
 
 	return OK;
 }
@@ -723,7 +792,7 @@ static int mp3_handler(request_rec *r) {
 
 	cfg->content_type = "audio/mpeg";      
 	info = create_request(r, cfg);
-	bank = mp3_create_content(r->pool, r->filename, r->uri, 0);
+	bank = mp3_create_content(r->pool, r->filename, r->uri, NULL, 0);
 	send_headers(r, cfg, info);
 
 	if ((status = stream_content(r, cfg, bank, info)) != OK)
@@ -742,7 +811,7 @@ static int ogg_handler(request_rec *r) {
 #ifdef DEBUG
 	printf("VALUE OF UDP %d\n", info->udp);
 #endif
-	bank = mp3_create_content(r->pool, r->filename, r->uri, 0);
+	bank = mp3_create_content(r->pool, r->filename, r->uri, NULL, 0);
 	send_headers(r, cfg, info);
 
 	if ((status = stream_content(r, cfg, bank, info)) != OK)
@@ -767,7 +836,7 @@ static int mp3_random_handler(request_rec *r) {
 	send_headers(r, cfg, info);
 
 	do {
-		while ((bank = cfg->dispatch->random(cfg->context, r->pool))) {
+		while ((bank = cfg->dispatch->each(cfg->context, r->pool, info->files, info->token, 1))) {
 			if (stream_content(r, cfg, bank, info) != OK)
 				return OK;
 		}
@@ -809,6 +878,7 @@ static int mp3_stream_handler(request_rec *r) {
 	}
 
 	while ((b = fgetc(file))){
+	while ((b = fgetc(file)) != EOF && (cfg->max_bytes <= 0 || x <= cfg->max_bytes) ) {
 		if (info->shout) {
 			if (shout_write(r,b, NULL, NULL, NULL, &latch) == -1) {
 				return HTTP_REQUEST_TIME_OUT;
@@ -856,6 +926,7 @@ static const command_rec mp3_module_cmds[] = {
 	{"MP3Loop", ap_set_flag_slot, (void *) XtOffsetOf(mp3_conf, enabled), OR_ALL, FLAG, MP3Loop},
 	{"MP3Random", ap_set_flag_slot, (void *) XtOffsetOf(mp3_conf, random_enabled), OR_ALL, FLAG, MP3Random},
 	{"MP3LimitPlayConnections", set_limit_connections, NULL, OR_ALL, TAKE1, MP3LimitPlayConnections},
+	{"MP3LimitBytesPerTrack", set_limit_bytespertrack, NULL, OR_ALL, TAKE1, MP3LimitBytesPerTrack},
 	{"MP3Cache", enable_cache, NULL, OR_ALL, FLAG, MP3Cache},
 	{"MP3", add_mp3, NULL, OR_ALL, TAKE1, MP3},
 	{"MP3PlayList", add_mp3_playlist, NULL, OR_ALL, TAKE1, MP3PlayList},
@@ -871,10 +942,18 @@ static const command_rec mp3_module_cmds[] = {
 	{"MP3Allow", add_mp3_accept, NULL, OR_ALL, TAKE1, MP3Allow},
 	{"MP3Deny", add_mp3_deny, NULL, OR_ALL, TAKE1, MP3Deny},
 	{"MP3DispatchAgent", add_dispatch_agent, NULL, OR_ALL, TAKE1, MP3DispatchAgent},
+#ifdef YP_ENABLED
+	{"MP3DirectoryServer", add_directory_server, NULL, OR_ALL, TAKE12, MP3DirectoryServer},
+#endif
 #ifdef MYSQL_ENABLED
 	{"MP3MySQLConnectInfo", mysql_add_connect_info, NULL, OR_ALL, TAKE3, MP3MySQLConnectInfo},
 	{"MP3MySQLInfo", mysql_add_database_info, NULL, OR_ALL, TAKE2, MP3MySQLInfo},
 	{"MP3MySQLTokenTable", mysql_add_token_table, NULL, OR_ALL, TAKE1, MP3MySQLTokenTable},
+#endif
+#ifdef PGSQL_ENABLED
+	{"MP3PgConnectInfo", pgsql_add_connect_info, NULL, OR_ALL, TAKE3, MP3MySQLConnectInfo},
+	{"MP3PgInfo", pgsql_add_database_info, NULL, OR_ALL, TAKE2, MP3MySQLInfo},
+	{"MP3PgTokenTable", pgsql_add_token_table, NULL, OR_ALL, TAKE1, MP3MySQLTokenTable},
 #endif
 	{NULL}
 };
@@ -884,6 +963,7 @@ static const handler_rec mp3_handlers[] = {
     { "mp3-file", mp3_handler }, 
     { "ogg-file", ogg_handler }, 
     { "mp3-play", mp3_play_handler }, 
+		{ "mp3-status", mp3_status_handler }, 
     { "mp3-random", mp3_random_handler }, 
 #ifdef SELECT_ENABLED
     { "mp3-selection", mp3_selection_handler }, 
@@ -922,7 +1002,7 @@ module MODULE_VAR_EXPORT mp3_module = {
 	mp3_fixup,             /* [#7] pre-run fixups                 */
 	NULL,                  /* [#9] log a transaction              */
 	NULL,                  /* [#2] header parser                  */
-	NULL,			             /* child_init                          */
+	mp3_child_init,        /* child_init                          */
 	NULL,					         /* child_exit                          */
 	mp3_request_init       /* [#0] post read-request              */
 #ifdef EAPI
