@@ -14,6 +14,8 @@ for (@array) {
 	s/^MP3_EXPORT\((.+)/$1/; 
 	s/\)//; 
 	s/ {$//; 
+	s/{
+$//; 
 	s/$/;/;
 	print $_;
 }
